/*
 * HSP.cpp
 *
 *  Created on: Feb 19, 2010
 *      Author: kohany
 */

#include "HSP.h"
#include <iomanip>
#include <stdlib.h>
#include <math.h>

namespace bio {

CMPVAL HSP::cmpby = TOPBEG;

HSP::HSP() :
   sim(0), pos(0), score(0) {
}
HSP::HSP(const HSP& other) :
   top(other.getTop()), bot(other.getBot()), sim(other.getSim()), pos(other.getPos()), score(other.getScore()) {
}
HSP::~HSP() {
}
HSP& HSP::operator =(const HSP& other) {
   top = other.getTop();
   bot = other.getBot();
   sim = other.getSim();
   pos = other.getPos();
   score = other.getScore();
   return *this;
}
HSP& HSP::operator +=(const HSP& other) {
   size_t oldLenTop = getTop().getLength();
   size_t oldLenBot = getBot().getLength();
   ssize_t ovlpTop = max((ssize_t) 0, (ssize_t)-dist(getTop(), other.getTop()));
   ssize_t ovlpBot = max((ssize_t) 0, (ssize_t)-dist(getBot(), other.getBot()));

   getTop() += other.getTop();
   getBot() += other.getBot();

   //setSim((interpolateScore(getSim(), oldLenTop, other.getSim(), other.getTop().getLength(), ovlpTop) + interpolateScore(getSim(), oldLenBot, other.getSim(), other.getBot().getLength(), ovlpBot)) / 2);
   setSim((getSim() * oldLenTop + other.getSim() * other.getTop().getLength()) / (oldLenTop + other.getTop().getLength()));
   //setPos((interpolateScore(getPos(), oldLenTop, other.getPos(), other.getTop().getLength(), ovlpTop) + interpolateScore(getPos(), oldLenBot, other.getPos(), other.getBot().getLength(), ovlpBot)) / 2);
   setPos((getPos() * oldLenTop + other.getPos() * other.getTop().getLength()) / (oldLenTop + other.getTop().getLength()));

   setScore((interpolateScore(getScore(), oldLenTop, other.getScore(), other.getTop().getLength(), ovlpTop) + interpolateScore(getScore(), oldLenBot, other.getScore(), other.getBot().getLength(),
         ovlpBot)) / 2);

   return *this;
}
HSP& HSP::operator -=(const HSP& other) throw (logic_error) {
   size_t oldLenTop = getTop().getLength();
   size_t oldLenBot = getBot().getLength();

   Fragment oldTop(getTop());
   getTop() -= other.getTop();
   double ratio = (double) getTop().getLength() / oldTop.getLength();
   double sign = getBot().isDirect() ? 1 : -1;
   if (oldTop.getBeg() == getTop().getBeg()) {
      getBot().assign(getBot().getRBeg(), getBot().getRBeg() + sign * ratio * getBot().getLength());
   } else {
      getBot().assign(getBot().getRBeg() + sign * (1.0 - ratio) * getBot().getLength(), getBot().getREnd());
   }

   setScore(((double) getScore() * getTop().getLength() / oldLenTop + (double) getScore() * getBot().getLength() / oldLenBot) / 2);

   return *this;
}
/**
 * Creates a pair of two HSPs which are parts of this HSP
 * flanking other HSP which has it's top fragment inside
 * top fragment of this.
 */
pair<HSP, HSP> HSP::subsplit(const HSP& other) const {
   pair<HSP, HSP> flanks(*this, *this);
   pair<Fragment, Fragment> topflanks(getTop().subsplit(other.getTop()));
   flanks.first.getTop() = topflanks.first;
   flanks.second.getTop() = topflanks.second;
   pair<double, double> ratios((double) flanks.first.getTop().getLength() / getTop().getLength(), (double) flanks.second.getTop().getLength() / getTop().getLength());
   flanks.first.getBot().assign(getBot().getBeg(), getBot().getBeg() + ratios.first * getBot().getLength());
   flanks.second.getBot().assign(getBot().getEnd() - ratios.second * getBot().getLength(), getBot().getEnd());
   flanks.first.setScore(ratios.first * getScore());
   flanks.second.setScore(ratios.second * getScore());
   return flanks;
}
const double HSP::getCmpVal() const {
   switch (HSP::cmpby) {
   case TOPBEG:
      return getTop().getBeg();
      break;
   case TOPEND:
      return getTop().getEnd();
      break;
   case BOTBEG:
      return getBot().getBeg();
      break;
   case BOTEND:
      return getBot().getEnd();
      break;
   case SIM:
      return getSim();
      break;
   case POS:
      return getPos();
      break;
   case SCORE:
   default:
      return getScore();
   }
}
const bool HSP::operator ==(const HSP& other) const {
   return getCmpVal() == other.getCmpVal();
}
const bool HSP::operator >(const HSP& other) const {
   return (getCmpVal() > other.getCmpVal());
}
const bool HSP::operator <(const HSP& other) const {
   return getCmpVal() < other.getCmpVal();
}
const bool HSP::operator >=(const HSP& other) const {
   return getCmpVal() >= other.getCmpVal();
}
const bool HSP::operator <=(const HSP& other) const {
   return getCmpVal() <= other.getCmpVal();
}
HSP& HSP::read(istream& is) {
   char dir;
   float sim, pos, score;
   is >> getTop() >> getBot() >> dir >> sim >> pos >> score;
   if (is.fail()) {
      // Clear failbit
      is.clear( is.rdstate() & (~istream::failbit));
      throw(istream::failure("ERROR: Failed to read HSP"));
   }
   if(dir == 'c') {
      getTop().setDirect(true);
      getBot().setDirect(false);
   }
   setSim(sim);
   setPos(pos);
   setScore(score);
   return *this;
}
void HSP::write(ostream& os) const {
   os << getTop() << getBot();
   os << setw(4) << (isDirect() ? 'd' : 'c');
   os << fixed << setw(7) << setprecision(4) << getSim() << " ";
   os << setw(7) << getPos() << " ";
   os << setprecision(0) << getScore() << " ";
}
//================================================================
istream& operator >>(istream& is, HSP& hsp) {
   string line;
   getline(is, line);
   // Unfortunatelly the following adds significant overhead since constructing istringstream is expensive.
   // As a solution HSP::read was added
   istringstream iss(line);
   try {
      hsp.read(iss);
   }
   catch(istream::failure err) {
      cerr << err.what() << endl;
      throw(istream::failure(string("ERROR: Failed to read line '")+line+"'"));
   }
   return is;
}
ostream& operator <<(ostream& os, const HSP& hsp) {
   hsp.write(os);
   return os;
}

bool areJoinable(const HSP& hsp1, const HSP& hsp2, float proxiRatio, int maxDist) {
   if (hsp1.isDirect() ^ hsp2.isDirect()) {
      return false;
   }
   if (areJoinable(hsp1.getTop(), hsp2.getTop(), maxDist) && areJoinable(hsp1.getBot(), hsp2.getBot(), maxDist)) {
      if (fabs( (double) range(hsp1.getTop(), hsp2.getTop(), false) / range(hsp1.getBot(), hsp2.getBot(), false) - 1) < proxiRatio) {
         if (areOrdered(hsp1.getTop(), hsp2.getTop(), false) && areOrdered(hsp1.getBot(), hsp2.getBot(), false)) {
            return true;
         }
         if (areOrdered(hsp2.getTop(), hsp1.getTop(), false) && areOrdered(hsp2.getBot(), hsp1.getBot(), false)) {
            return true;
         }
      }
   }
   return false;
}

double interpolateScore(double score1, double ctr1, double score2, double ctr2, double shared) {
//   return score1 + score2 - min(shared*score1/ctr1, shared*score2/ctr2);
   return score1 + score2 - (shared*score1/ctr1 + shared*score2/ctr2) / 2;
}
};
