/***************************************************************************
 cdb.cpp  -  description
 -------------------
 begin                : Thu Jan 23 2003
 copyright            : (C) 2003 by Oleksiy Kohany
 email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *                                                                         *
 *   class CDb - implements basic operations on multiple sequence file:    *
 *   - searching particular locus                                          *
 *   - extracting subsequence from lucus rapidly, assumes that             *
 *     all data lines are of the same length                               *
 *                                                                         *
 ***************************************************************************/
#include "cdb.h"
//#include "constants.h"
#include <iostream>
#include <string.h>
#include <sstream>
using namespace std;
//#define DEBUG

#include <sys/types.h>
#include <sys/stat.h>

namespace bio {
ostream& operator <<(ostream& os, CLP& lp) {
   os << lp.fname << "\t" << lp.headPos << "\t" << lp.dataPos << "\n";
   return os;
}

/** ==================================================== */
CDb::CDb(const char* fn) {
   ifs = 0;
   if (fn == NULL) {
      throw invalid_argument("File name is NULL");
   }
   ifs = new ifstream;
   sDbName = sDbDir = fn;
   sDbDir.erase(sDbDir.rfind('/') + 1);
   sDbName.erase(0, sDbDir.rfind('/') + 1);
   if (loadIdx(fn)) {
#ifdef DEBUG
      cerr << "close_ifs("<<sDbName<<") in CDb::CDb("<<fn<<");\n";
#endif
      sIDX = fn;
   } else {
      open_ifs(sDbName.c_str());
      sIDX.assign(fn);
      sIDX.append(".idx");
      int size = loadIndexUpToDate(fn, sIDX.c_str());
#ifdef DEBUG
      cerr << "loadIndexUpToDate(" << fn << ", " << sIDX << ") = " << size << "\n";
#endif
   }
   iIDXSize = lpmap.size();
   iLastPos = 0;
}
/** ==================================================== */
CDb::~CDb() {
#ifdef DEBUG
   cerr << "close_ifs("<<sDbName<<") in CDb::~CDb();\n";
#endif
   close_ifs();
   delete ifs;
   if (lpmap.size() > 0 && iIDXSize < lpmap.size()) {
      saveIdx(sIDX.c_str());
   }
}
/** ==================================================== */
size_t CDb::getHeadPos(const char* locus) {
   findLocus(locus);
   return lpmap[locus].headPos;
}
/** ==================================================== */
size_t CDb::getDataPos(const char* locus) {
   findLocus(locus);
   return lpmap[locus].dataPos;
}
/** ==================================================== */
istream& CDb::getHead(const char* locus) {
   findLocus(locus);
   ifs->clear();
   ifs->seekg(lpmap[locus].headPos, ios::beg);
   return *ifs;
}
/** ==================================================== */
istream& CDb::getData(const char* locus) {
   findLocus(locus);
   ifs->clear();
   ifs->seekg(lpmap[locus].dataPos, ios::beg);
   return *ifs;
}
/** ==================================================== */
void CDb::findLocus(const char* locus) {
   if (lpmap.find(locus) != lpmap.end()) {
      if (sDbName != lpmap[locus].fname) {
#ifdef DEBUG
         cerr << "close_ifs("<<sDbName<<") in int CDb::findLocus("<<locus<<"); \n";
#endif
         close_ifs();
         open_ifs(lpmap[locus].fname.c_str());
         iLastPos = 0;
      }
      if (iIDXSize == 0) {
         return;
      }
      if (cmpage((sDbDir + lpmap[locus].fname).c_str(), sIDX.c_str()) > 0) {
#ifdef DEBUG
         cerr << sIDX << " is older than " << lpmap[locus].fname << "\n";
#endif
      } else {
#ifdef DEBUG
         cerr << sIDX << " is younger than " << lpmap[locus].fname << "\n";
#endif
         return;
      }
   }
#ifdef DEBUG
   cerr << "IOSTREAM: '" << locus << "' Not found in index. Tracking. iLastPos= " << iLastPos << "\n";
   int iCnt = 0;
#endif
   size_t iHP;
   size_t iDP;
   iHP = iLastPos;
   iDP = iLastPos;
   char sLine[MAXLINELENGTH];
   //   string sLine;
   char sName[MAXLINELENGTH];
   if (ifs->is_open()) {
      ifs->clear();
      ifs->seekg(iLastPos, ios::beg);
      while (!ifs->getline(sLine, MAXLINELENGTH - 1).eof()) {
         //      while(getline(*ifs, sLine) && !ifs->fail()) {
#ifdef DEBUG
         cerr << "'" << sLine << "'\n";
#endif
         ifs->clear();
         sLine[MAXLINELENGTH - 1] = '\0';
         if (sLine[0] == ';' || sLine[0] == '>') {
            switch (sLine[0]) {
            case '>':
               sscanf(sLine + 1, "%s", sName);
               break;
            case ';':
               while (!ifs->getline(sLine, MAXLINELENGTH).eof() && sLine[0] == ';') {
                  ifs->clear();
               }
               //                  while(getline(*ifs, sLine) && !ifs->fail() && sLine[0] == ';');
               sscanf(sLine, "%s", sName);
               break;
            }
            iLastPos = iDP = ifs->tellg();
            ifs->clear();
            lpmap[sName].assign(sDbName, iHP, iDP);
#ifdef DEBUG
            iCnt++;
            cerr << sName << "<=>" << locus << "\n";
#endif
            if (strcmp(sName, locus) == 0) {
               return;
            }
         }
         iHP = iLastPos = ifs->tellg();
      }
#ifdef DEBUG
      cerr << "Failed ? " << ifs->fail() << "\n";
      cerr << "'" << sLine << "'\n";
      cerr << "Scanned " << iCnt << " locies\n";
#endif
   }
   ifs->clear();
   ostringstream oss;
   oss << __FILE__ << ":" << __LINE__ << ": Locus " << locus << " not found";
   throw out_of_range(oss.str());
}
/** ==================================================== */
int CDb::loadIdx(const char *fn) {
   FILE *fidx;
   char sLine[MAXLINELENGTH];
   char sLName[MAXNAMESIZE];
   char sFName[MAXNAMESIZE];
   size_t nCom = 0, nFrom = 0;
   lpmap.clear();
   if ((fidx = fopen(fn, "r")) == 0) {
      return 0;
   }
   while (fgets(sLine, MAXLINELENGTH, fidx) != 0) {
      if (sscanf(sLine, "%s%s%ld%ld", sFName, sLName, &nCom, &nFrom) < 4) {
         lpmap.clear();
         break;
      }
      lpmap[sLName].assign(sFName, nCom, nFrom);
   }
   fclose(fidx);
   return lpmap.size();
}
/** ==================================================== */
void CDb::saveIdx(const char* fn) {
   ofstream ofs(fn);
   for (LPMAP::iterator mptr = lpmap.begin(); mptr != lpmap.end(); mptr++) {
      ofs << (*mptr).second.fname << "\t" << (*mptr).first << "\t" << (*mptr).second.headPos << "\t" << (*mptr).second.dataPos << endl;
   }
   ofs.close();
}
/** ==================================================== */
void CDb::saveIdx(ostream& os) {
   for (LPMAP::iterator mptr = lpmap.begin(); mptr != lpmap.end(); mptr++) {
      os << sDbDir << (*mptr).second.fname << "\t" << (*mptr).first << "\t" << (*mptr).second.headPos << "\t" << (*mptr).second.dataPos << endl;
   }
   os.flush();
}
/** ==================================================== */
void CDb::open_ifs(const string& fn) {
   if(fn.size() == 0) {
      throw runtime_error("Empty file name");
   }
   if (sDbName != fn) {
      sDbName.assign(fn);
   }
   string fname(sDbDir);
   fname.append(sDbName);
   ifs->clear();
   ifs->open(fname.c_str());
   if (!ifs->is_open()) {
      sDbName.assign("");
      cerr << "Failed to open " << fname.c_str() << " for reading" << endl;
      throw runtime_error("Failed to open file");
   }
   sDbName = fn;
#ifdef DEBUG
   cerr << "Opened " << sDbName << " as " << fname << "\n";
#endif
}
/** ==================================================== */
void CDb::close_ifs() {
#ifdef DEBUG
   cerr << "Closing " << sDbName << "\n";
#endif
   sDbName.assign("");
   if (ifs->is_open()) {
      ifs->close();
   }
}
/** ==================================================== */
void CDb::rewind_ifs() {
   if (ifs->is_open()) {
      ifs->seekg(0, ios::beg);
   }
}
/** ==================================================== */
bool CDb::exist(const char *pcFName) {
   FILE *FIN;
   if ((FIN = fopen(pcFName, "r")) == 0) {
      return false;
   }
   fclose(FIN);
   return true;
}
/** ==============================================================
 Reads a fragment of specified locus into sq
 Returns number of symbols actually written into sq
 ============================================================== */
size_t CDb::fragment(const char* pcName, ssize_t iF, ssize_t iT, BioSubSequence& sq, bool bGap) {
   if (iF <= 0) {
      iF = 1;
   }
   if (iT <= 0) {
      iT = 1;
   }
   sq.clear();
   //Reading header of the source sequence to get it's position if available.
   try {
      sq.readHead(getHead(pcName));
      getData(pcName);
   }
   catch(out_of_range err) {
      cerr << err.what() << endl;
      return 0;
   }
   if (iF > iT) {
      sq.adjust_position(iT, iF - fragment(iT, iF, sq, bGap));
      sq.complement();
   } else {
      sq.adjust_position(iF, iT - fragment(iF, iT, sq, bGap));
   }
   return sq.size();
}
#ifdef _USE_NAMEDSUBSEQUENCE_
/** ==============================================================
 Reads a fragment of specified locus into sq
 Returns number of symbols actually written into sq
 ============================================================== */
size_t CDb::fragment(Fragment frag, NamedSubSequence& sq, bool bGap) {
   if (frag.getBeg() < 0) {
      frag.setBeg(0);
   }
   if (frag.getEnd() <= 0) {
      frag.setEnd(1);
   }

   sq.clear();
   //Reading header of the source sequence to get it's position if available.
   try {
      sq.readHeader(getHead(frag.getName().c_str()));
      getData(frag.getName().c_str());
   }
   catch(out_of_range err) {
      cerr << err.what() << endl;
      return 0;
   }
   frag.setEnd(frag.getEnd() - fragment(frag.getBeg() + 1, frag.getEnd(), sq, bGap));
   sq.setFragment(frag.getBeg(), frag.getEnd());
   if (!frag.isDirect()) {
      complement(sq.begin(), sq.end());
   }
   return sq.size();
}
#endif

/** ==============================================================
 Calculaes starting position of the fragment and reads it.
 Returns difference between what should be read and what actually was read.
 ============================================================== */
size_t CDb::fragment(size_t iFr, size_t iTo, string& sq, bool bGap) {
   size_t iF = min(iFr, iTo);
   size_t iT = max(iFr, iTo);
   size_t iW = getWidth();
   if (iW == 0) {
      cerr << "CDb::getWidth() returned 0" << endl;
      return iT - iF + 1;
   }
   int iRP = iF + (iF - 1) / iW - 1;
   ifs->seekg(iRP, ios_base::cur);
   return fragment(ifs, iT - iF + 1, sq, bGap);
}
/** ==============================================================
 This function reads iCnt letters from ifs to sq, or untill ';' or '>' is met
 Returns difference between what should be read and what actually was read.
 ============================================================== */
size_t CDb::fragment(ifstream *ifs, size_t iCnt, string& sq, bool bGap) {
   string line;
   ifs->clear();
   sq.assign("");
   sq.reserve(iCnt);
   while (!getline(*ifs, line).fail() && line[0] != ';' && line[0] != '>') {
      for (string::iterator sptr = line.begin(); *sptr != '\n' && *sptr != '\0'; sptr++) {
         if ((!bGap && (isalpha(*sptr) || *sptr == '*')) || (bGap && isgraph(*sptr))) {
            sq.append(1, *sptr);
            if (--iCnt <= 0) {
               return 0;
            }
         }
      }
   }
   return iCnt;
}
/** ==============================================================
 Estimates length of data lines in sequence file.
 The whole algorithm assumes that all lines in the sequence file are of the same length
 Returns astimated length of datalines (width of the file)
 ============================================================= */
size_t CDb::getWidth() {
   size_t iPos = ifs->tellg();
   ifs->clear();
   size_t iWf = 0;
   size_t iWt = 0;
   char c;
   while (ifs->get(c).good()) {
      if (c == ';' || c == '>') {
         break;
      }
      if (c == '\n' && iWt > 0) {
         if (iWf == iWt) {
            break;
         }
         if (iWf == 0) {
            iWf = iWt;
            iWt = 0;
            continue;
         }
         if (iWf != iWt) {
            //fprintf(stderr, "CDb::getWidth(): First 2 lines are of different length: %ld - %ld\n", iWf, iWt);
            //fprintf(stderr, "Possible reasons: short sequence (less than 2 lines), outdated index or bad sequence format.\n");
         }
         if (iWf < iWt) {
            iWf = iWt;
            break;
         }
      } else {
         iWt++;
      }
   }
   ifs->clear();
   ifs->seekg(iPos, ios::beg);
   return iWf;
}
/** No descriptions */
int CDb::loadIndexUpToDate(const char* db_name, const char* idx_name) {
   if (db_name == 0) {
      return 0;
   }
   if (cmpage(db_name, idx_name) > 0) {
      fprintf(stderr, "Index file %s is older than %s\n", idx_name, db_name);
      return 0;
   }
   return loadIdx(idx_name);
}
/** No descriptions */
int cmpage(const char* name1, const char* name2) {
   if (name1 == 0 || name2 == 0) {
      return 0;
   }
   struct stat stat1;
   struct stat stat2;
   if (stat(name1, &stat1)) {
#ifdef DEBUG
      cerr << "::stat(" << name1 << ", &stat1); failed\n";
#endif
      return 0;
   }
   if (stat(name2, &stat2)) {
#ifdef DEBUG
      cerr << "::stat(" << name2 << ", &stat2); failed\n";
#endif
      return 0;
   }
#ifdef DEBUG
   cerr << name1 << " : mtime = " << stat1.st_mtime << " : ctime = " << stat1.st_ctime << " : atime = " << stat1.st_atime << "\n";
   cerr << name2 << " : mtime = " << stat2.st_mtime << " : ctime = " << stat2.st_ctime << " : atime = " << stat2.st_atime << "\n";
#endif
   return stat1.st_mtime - stat2.st_mtime;
}
};
